# License: MIT
# Copyright © 2022 Frequenz Energy-as-a-Service GmbH

"""A logical meter for calculating high level metrics for a microgrid."""

from ._logical_meter import LogicalMeter

__all__ = ["LogicalMeter"]
