# License: MIT
# Copyright © 2022 Frequenz Energy-as-a-Service GmbH

"""Formula Engine Exceptions."""


class FormulaEngineError(Exception):
    """An error occurred while fetching metrics or applying the formula on them."""
