# License: MIT
# Copyright © 2022 Frequenz Energy-as-a-Service GmbH

"""Utility types and functions for internal use."""

from ._singleton_meta import SingletonMeta

# Explicitly declare the public API.
__all__ = ["SingletonMeta"]
